@@ -1,21 +1,32 @@
 """
 Useful non-core functionality, e.g. functions composing multiple operations.
 """
+from __future__ import with_statement
 
 from os import getcwd, sep
 import os.path
 from datetime import datetime
 from tempfile import mkdtemp
 
-from fabric.network import needs_host
-from fabric.operations import local, run, put
+from fabric.network import needs_host, key_filenames, normalize
+from fabric.operations import local, run, sudo, put
 from fabric.state import env, output
+from fabric.context_managers import cd
 
 __all__ = ['rsync_project', 'upload_project']
 
 @needs_host
-def rsync_project(remote_dir, local_dir=None, exclude=(), delete=False,
-    extra_opts=''):
+def rsync_project(
+    remote_dir,
+    local_dir=None,
+    exclude=(),
+    delete=False,
+    extra_opts='',
+    ssh_opts='',
+    capture=False,
+    upload=True,
+    default_opts='-pthrvz'
+):
     """
     Synchronize a remote directory with the current project directory via rsync.
 
@@ -41,7 +52,7 @@ def rsync_project(remote_dir, local_dir=None, exclude=(), delete=False,
 
         * If ``local_dir`` ends with a trailing slash, the files will be
           dropped inside of ``remote_dir``. E.g.
-          ``rsync_project("/home/username/project", "foldername/")`` will drop
+          ``rsync_project("/home/username/project/", "foldername/")`` will drop
           the contents of ``foldername`` inside of ``/home/username/project``.
         * If ``local_dir`` does **not** end with a trailing slash (and this
           includes the default scenario, when ``local_dir`` is not specified),
@@ -63,6 +74,13 @@ def rsync_project(remote_dir, local_dir=None, exclude=(), delete=False,
       longer exist locally. Defaults to False.
     * ``extra_opts``: an optional, arbitrary string which you may use to pass
       custom arguments or options to ``rsync``.
+    * ``ssh_opts``: Like ``extra_opts`` but specifically for the SSH options
+      string (rsync's ``--rsh`` flag.)
+    * ``capture``: Sent directly into an inner `~fabric.operations.local` call.
+    * ``upload``: a boolean controlling whether file synchronization is
+      performed up or downstream. Upstream by default.
+    * ``default_opts``: the default rsync options ``-pthrvz``, override if
+      desired (e.g. to remove verbosity, etc).
 
     Furthermore, this function transparently honors Fabric's port and SSH key
     settings. Calling this function when the current host string contains a
@@ -73,8 +91,14 @@ def rsync_project(remote_dir, local_dir=None, exclude=(), delete=False,
     constructed by this function is the following::
 
         rsync [--delete] [--exclude exclude[0][, --exclude[1][, ...]]] \\
-            -pthrvz [extra_opts] <local_dir> <host_string>:<remote_dir>
+            [default_opts] [extra_opts] <local_dir> <host_string>:<remote_dir>
 
+    .. versionadded:: 1.4.0
+        The ``ssh_opts`` keyword argument.
+    .. versionadded:: 1.4.1
+        The ``capture`` keyword argument.
+    .. versionadded:: 1.8.0
+        The ``default_opts`` keyword argument.
     """
     # Turn single-string exclude into a one-item list for consistency
     if not hasattr(exclude, '__iter__'):
@@ -85,48 +109,61 @@ def rsync_project(remote_dir, local_dir=None, exclude=(), delete=False,
     exclusions = tuple([str(s).replace('"', '\\\\"') for s in exclude])
     # Honor SSH key(s)
     key_string = ""
-    if env.key_filename:
-        keys = env.key_filename
-        # For ease of use, coerce stringish key filename into list
-        if not isinstance(env.key_filename, (list, tuple)):
-            keys = [keys]
+    keys = key_filenames()
+    if keys:
         key_string = "-i " + " -i ".join(keys)
-    # Honor nonstandard port
-    port_string = ("-p %s" % env.port) if (env.port != '22') else ""
+    # Port
+    user, host, port = normalize(env.host_string)
+    port_string = "-p %s" % port
     # RSH
     rsh_string = ""
-    if key_string or port_string:
-        rsh_string = "--rsh='ssh %s %s'" % (port_string, key_string)
+    rsh_parts = [key_string, port_string, ssh_opts]
+    if any(rsh_parts):
+        rsh_string = "--rsh='ssh %s'" % " ".join(rsh_parts)
     # Set up options part of string
     options_map = {
         'delete': '--delete' if delete else '',
         'exclude': exclude_opts % exclusions,
         'rsh': rsh_string,
-        'extra': extra_opts
+        'default': default_opts,
+        'extra': extra_opts,
     }
-    options = "%(delete)s%(exclude)s -pthrvz %(extra)s %(rsh)s" % options_map
+    options = "%(delete)s%(exclude)s %(default)s %(extra)s %(rsh)s" % options_map
     # Get local directory
     if local_dir is None:
         local_dir = '../' + getcwd().split(sep)[-1]
     # Create and run final command string
-    cmd = "rsync %s %s %s@%s:%s" % (options, local_dir, env.user,
-        env.host, remote_dir)
+    if host.count(':') > 1:
+        # Square brackets are mandatory for IPv6 rsync address,
+        # even if port number is not specified
+        remote_prefix = "[%s@%s]" % (user, host)
+    else:
+        remote_prefix = "%s@%s" % (user, host)
+    if upload:
+        cmd = "rsync %s %s %s:%s" % (options, local_dir, remote_prefix, remote_dir)
+    else:
+        cmd = "rsync %s %s:%s %s" % (options, remote_prefix, remote_dir, local_dir)
+
     if output.running:
         print("[%s] rsync_project: %s" % (env.host_string, cmd))
-    return local(cmd)
+    return local(cmd, capture=capture)
 
 
-def upload_project(local_dir=None, remote_dir=""):
+def upload_project(local_dir=None, remote_dir="", use_sudo=False):
     """
     Upload the current project to a remote system via ``tar``/``gzip``.
 
     ``local_dir`` specifies the local project directory to upload, and defaults
     to the current working directory.
-    
+
     ``remote_dir`` specifies the target directory to upload into (meaning that
     a copy of ``local_dir`` will appear as a subdirectory of ``remote_dir``)
     and defaults to the remote user's home directory.
 
+    ``use_sudo`` specifies which method should be used when executing commands
+    remotely. ``sudo`` will be used if use_sudo is True, otherwise ``run`` will
+    be used.
+
     This function makes use of the ``tar`` and ``gzip`` programs/libraries,
     thus it will not work too well on Win32 systems unless one is using Cygwin
     or something similar. It will attempt to clean up the local and remote
@@ -134,7 +171,12 @@ def upload_project(local_dir=None, remote_dir=""):
 
     .. versionchanged:: 1.1
         Added the ``local_dir`` and ``remote_dir`` kwargs.
+
+    .. versionchanged:: 1.7
+        Added the ``use_sudo`` kwarg.
     """
+    runner = use_sudo and sudo or run
+
     local_dir = local_dir or os.getcwd()
 
     # Remove final '/' in local_dir so that basename() works
@@ -148,10 +190,11 @@ def upload_project(local_dir=None, remote_dir=""):
     try:
         tar_path = os.path.join(tmp_folder, tar_file)
         local("tar -czf %s -C %s %s" % (tar_path, local_path, local_name))
-        put(tar_path, target_tar)
-        try:
-            run("tar -xzf %s" % tar_file)
-        finally:
-            run("rm -f %s" % tar_file)
+        put(tar_path, target_tar, use_sudo=use_sudo)
+        with cd(remote_dir):
+            try:
+                runner("tar -xzf %s" % tar_file)
+            finally:
+                runner("rm -f %s" % tar_file)
     finally:
         local("rm -rf %s" % tmp_folder)
