@@ -8,30 +8,33 @@ fabfile, and executing the commands given.
 The other callables defined in this module are internal only. Anything useful
 to individuals leveraging Fabric as a library, should be kept elsewhere.
 """
-
-from collections import defaultdict
-from operator import add, isMappingType
+import getpass
+from operator import isMappingType
 from optparse import OptionParser
 import os
 import sys
 import types
 
-from fabric import api, state  # For checking callables against the API, & easy mocking
-from fabric.contrib import console, files, project  # Ditto
-from fabric.network import denormalize, disconnect_all
+# For checking callables against the API, & easy mocking
+from fabric import api, state, colors
+from fabric.contrib import console, files, project
+
+from fabric.network import disconnect_all, ssh
 from fabric.state import env_options
-from fabric.tasks import Task, execute
+from fabric.tasks import Task, execute, get_task_details
 from fabric.task_utils import _Dict, crawl
-from fabric.utils import abort, indent
+from fabric.utils import abort, indent, warn, _pty_size
+
 
 # One-time calculation of "all internal callables" to avoid doing this on every
 # check of a given fabfile callable (in is_classic_task()).
-_modules = [api, project, files, console]
+_modules = [api, project, files, console, colors]
 _internals = reduce(lambda x, y: x + filter(callable, vars(y).values()),
     _modules,
     []
 )
 
+
 # Module recursion cache
 class _ModuleCache(object):
     """
@@ -56,7 +59,7 @@ def load_settings(path):
     """
     Take given file path and return dictionary of any key=value pairs found.
 
-    Usage docs are in docs/usage/fab.rst, in "Settings files."
+    Usage docs are in sites/docs/usage/fab.rst, in "Settings files."
     """
     if os.path.exists(path):
         comments = lambda s: s and not s.startswith("#")
@@ -77,14 +80,15 @@ def _is_package(path):
     )
 
 
-def find_fabfile():
+def find_fabfile(names=None):
     """
     Attempt to locate a fabfile, either explicitly or by searching parent dirs.
 
-    Usage docs are in docs/usage/fabfiles.rst, in "Fabfile discovery."
+    Usage docs are in sites/docs/usage/fabfiles.rst, in "Fabfile discovery."
     """
-    # Obtain env value
-    names = [state.env.fabfile]
+    # Obtain env value if not given specifically
+    if names is None:
+        names = [state.env.fabfile]
     # Create .py version if necessary
     if not names[0].endswith('.py'):
         names += [names[0] + '.py']
@@ -116,11 +120,16 @@ def is_classic_task(tup):
     Takes (name, object) tuple, returns True if it's a non-Fab public callable.
     """
     name, func = tup
-    return (
-        callable(func)
-        and (func not in _internals)
-        and not name.startswith('_')
-    )
+    try:
+        is_classic = (
+            callable(func)
+            and (func not in _internals)
+            and not name.startswith('_')
+        )
+    # Handle poorly behaved __eq__ implementations
+    except (ValueError, TypeError):
+        is_classic = False
+    return is_classic
 
 
 def load_fabfile(path, importer=None):
@@ -201,8 +210,12 @@ def extract_tasks(imported_vars):
         name, obj = tup
         if is_task_object(obj):
             state.env.new_style_tasks = True
-            # Honor instance.name
-            new_style_tasks[obj.name] = obj
+            # Use instance.name if defined
+            if obj.name and obj.name != 'undefined':
+                new_style_tasks[obj.name] = obj
+            else:
+                obj.name = name
+                new_style_tasks[name] = obj
             # Handle aliasing
             if obj.aliases is not None:
                 for alias in obj.aliases:
@@ -229,7 +242,7 @@ def is_task_module(a):
     """
     #return (type(a) is types.ModuleType and
     #        any(map(is_task_object, vars(a).values())))
-    if type(a) is types.ModuleType and a not in _seen:
+    if isinstance(a, types.ModuleType) and a not in _seen:
         # Flag module as seen
         _seen.add(a)
         # Signal that we need to check it out
@@ -256,7 +269,9 @@ def parse_options():
     # Initialize
     #
 
-    parser = OptionParser(usage="fab [options] <command>[:arg1,arg2=val2,host=foo,hosts='h1;h2',...] ...")
+    parser = OptionParser(
+        usage=("fab [options] <command>"
+               "[:arg1,arg2=val2,host=foo,hosts='h1;h2',...] ..."))
 
     #
     # Define options that don't become `env` vars (typically ones which cause
@@ -264,13 +279,25 @@ def parse_options():
     # --version)
     #
 
-    # Version number (optparse gives you --version but we have to do it
-    # ourselves to get -V too. sigh)
-    parser.add_option('-V', '--version',
+    # Display info about a specific command
+    parser.add_option('-d', '--display',
+        metavar='NAME',
+        help="print detailed info about command NAME"
+    )
+
+    # Control behavior of --list
+    LIST_FORMAT_OPTIONS = ('short', 'normal', 'nested')
+    parser.add_option('-F', '--list-format',
+        choices=LIST_FORMAT_OPTIONS,
+        default='normal',
+        metavar='FORMAT',
+        help="formats --list, choices: %s" % ", ".join(LIST_FORMAT_OPTIONS)
+    )
+
+    parser.add_option('-I', '--initial-password-prompt',
         action='store_true',
-        dest='show_version',
         default=False,
-        help="show program's version number and exit"
+        help="Force password prompt up-front"
     )
 
     # List Fab commands found in loaded fabfiles/source files
@@ -281,6 +308,14 @@ def parse_options():
         help="print list of possible commands and exit"
     )
 
+    # Allow setting of arbitrary env vars at runtime.
+    parser.add_option('--set',
+        metavar="KEY=VALUE,...",
+        dest='env_settings',
+        default="",
+        help="comma separated KEY=VALUE pairs to set Fab env vars"
+    )
+
     # Like --list, but text processing friendly
     parser.add_option('--shortlist',
         action='store_true',
@@ -289,18 +324,13 @@ def parse_options():
         help="alias for -F short --list"
     )
 
-    # Control behavior of --list
-    LIST_FORMAT_OPTIONS = ('short', 'normal', 'nested')
-    parser.add_option('-F', '--list-format',
-        choices=LIST_FORMAT_OPTIONS,
-        default='normal',
-        help="formats --list, choices: %s" % ", ".join(LIST_FORMAT_OPTIONS)
-    )
-
-    # Display info about a specific command
-    parser.add_option('-d', '--display',
-        metavar='COMMAND',
-        help="print detailed info about a given command and exit"
+    # Version number (optparse gives you --version but we have to do it
+    # ourselves to get -V too. sigh)
+    parser.add_option('-V', '--version',
+        action='store_true',
+        dest='show_version',
+        default=False,
+        help="show program's version number and exit"
     )
 
     #
@@ -318,12 +348,14 @@ def parse_options():
     opts, args = parser.parse_args()
     return parser, opts, args
 
+
 def _is_task(name, value):
     """
     Is the object a task as opposed to e.g. a dict or int?
     """
     return is_classic_task((name, value)) or is_task_object(value)
 
+
 def _sift_tasks(mapping):
     tasks, collections = [], []
     for name, value in mapping.iteritems():
@@ -335,6 +367,7 @@ def _sift_tasks(mapping):
     collections = sorted(collections)
     return tasks, collections
 
+
 def _task_names(mapping):
     """
     Flatten & sort task names in a breadth-first fashion.
@@ -351,11 +384,12 @@ def _task_names(mapping):
         tasks.extend(map(join, _task_names(module)))
     return tasks
 
+
 def _print_docstring(docstrings, name):
     if not docstrings:
         return False
     docstring = crawl(name, state.commands).__doc__
-    if type(docstring) in types.StringTypes:
+    if isinstance(docstring, basestring):
         return docstring
 
 
@@ -366,6 +400,7 @@ def _normal_list(docstrings=True):
     max_len = reduce(lambda a, b: max(a, len(b)), task_names, 0)
     sep = '  '
     trail = '...'
+    max_width = _pty_size()[1] - 1 - len(trail)
     for name in task_names:
         output = None
         docstring = _print_docstring(docstrings, name)
@@ -373,7 +408,7 @@ def _normal_list(docstrings=True):
             lines = filter(None, docstring.splitlines())
             first_line = lines[0].strip()
             # Truncate it if it's longer than N chars
-            size = 75 - (max_len + len(sep) + len(trail))
+            size = max_width - (max_len + len(sep) + len(trail))
             if len(first_line) > size:
                 first_line = first_line[:size] + trail
             output = name.ljust(max_len) + sep + first_line
@@ -400,6 +435,7 @@ def _nested_list(mapping, level=1):
 COMMANDS_HEADER = "Available commands"
 NESTED_REMINDER = " (remember to call as module.[...].task)"
 
+
 def list_commands(docstring, format_):
     """
     Print all found commands/tasks, then exit. Invoked with ``-l/--list.``
@@ -437,10 +473,14 @@ def display_command(name):
         msg = "Task '%s' does not appear to exist. Valid task names:\n%s"
         abort(msg % (name, "\n".join(_normal_list(False))))
     # Print out nicely presented docstring if found
-    if command.__doc__:
+    if hasattr(command, '__details__'):
+        task_details = command.__details__()
+    else:
+        task_details = get_task_details(command)
+    if task_details:
         print("Displaying detailed information for task '%s':" % name)
         print('')
-        print(indent(command.__doc__, strip=True))
+        print(indent(task_details, strip=True))
         print('')
     # Or print notice if not
     else:
@@ -479,7 +519,7 @@ def parse_arguments(arguments):
     """
     Parse string list into list of tuples: command, args, kwargs, hosts, roles.
 
-    See docs/usage/fab.rst, section on "per-task arguments" for details.
+    See sites/docs/usage/fab.rst, section on "per-task arguments" for details.
     """
     cmds = []
     for cmd in arguments:
@@ -491,8 +531,9 @@ def parse_arguments(arguments):
         if ':' in cmd:
             cmd, argstr = cmd.split(':', 1)
             for pair in _escape_split(',', argstr):
-                k, _, v = pair.partition('=')
-                if _:
+                result = _escape_split('=', pair)
+                if len(result) > 1:
+                    k, v = result
                     # Catch, interpret host/hosts/role/roles/exclude_hosts
                     # kwargs
                     if k in ['host', 'hosts', 'role', 'roles', 'exclude_hosts']:
@@ -510,7 +551,7 @@ def parse_arguments(arguments):
                     else:
                         kwargs[k] = v
                 else:
-                    args.append(k)
+                    args.append(result[0])
         cmds.append((cmd, args, kwargs, hosts, roles, exclude_hosts))
     return cmds
 
@@ -539,10 +580,12 @@ def update_output_levels(show, hide):
             state.output[key] = False
 
 
-from fabric.tasks import _parallel_tasks
+def show_commands(docstring, format, code=0):
+    print("\n".join(list_commands(docstring, format)))
+    sys.exit(code)
 
 
-def main():
+def main(fabfile_locations=None):
     """
     Main command-line execution loop.
     """
@@ -554,6 +597,22 @@ def main():
         arguments = parser.largs
         remainder_arguments = parser.rargs
 
+        # Allow setting of arbitrary env keys.
+        # This comes *before* the "specific" env_options so that those may
+        # override these ones. Specific should override generic, if somebody
+        # was silly enough to specify the same key in both places.
+        # E.g. "fab --set shell=foo --shell=bar" should have env.shell set to
+        # 'bar', not 'foo'.
+        for pair in _escape_split(',', options.env_settings):
+            pair = _escape_split('=', pair)
+            # "--set x" => set env.x to True
+            # "--set x=" => set env.x to ""
+            key = pair[0]
+            value = True
+            if len(pair) == 2:
+                value = pair[1]
+            state.env[key] = value
+
         # Update env with any overridden option values
         # NOTE: This needs to remain the first thing that occurs
         # post-parsing, since so many things hinge on the values in env.
@@ -566,19 +625,23 @@ def main():
             if key in state.env and isinstance(state.env[key], basestring):
                 state.env[key] = state.env[key].split(',')
 
+        # Feed the env.tasks : tasks that are asked to be executed.
+        state.env['tasks'] = arguments
+
         # Handle output control level show/hide
         update_output_levels(show=options.show, hide=options.hide)
 
         # Handle version number option
         if options.show_version:
             print("Fabric %s" % state.env.version)
+            print("Paramiko %s" % ssh.__version__)
             sys.exit(0)
 
         # Load settings from user settings file, into shared env dict.
         state.env.update(load_settings(state.env.rcfile))
 
         # Find local fabfile path or abort
-        fabfile = find_fabfile()
+        fabfile = find_fabfile(fabfile_locations)
         if not fabfile and not remainder_arguments:
             abort("""Couldn't find any fabfiles!
 
@@ -590,6 +653,7 @@ Remember that -f can be used to specify fabfile path, and use -h for help.""")
         # Load fabfile (which calls its module-level code, including
         # tweaks to env values) and put its commands in the shared commands
         # dict
+        default = None
         if fabfile:
             docstring, callables, default = load_fabfile(fabfile)
             state.commands.update(callables)
@@ -621,8 +685,7 @@ Remember that -f can be used to specify fabfile path, and use -h for help.""")
 
         # List available commands
         if options.list_commands:
-            print("\n".join(list_commands(docstring, options.list_format)))
-            sys.exit(0)
+            show_commands(docstring, options.list_format)
 
         # Handle show (command-specific help) option
         if options.display:
@@ -646,9 +709,10 @@ Remember that -f can be used to specify fabfile path, and use -h for help.""")
                 unknown_commands.append(tup[0])
 
         # Abort if any unknown commands were specified
-        if unknown_commands:
-            abort("Command(s) not found:\n%s" \
+        if unknown_commands and not state.env.get('skip_unknown_tasks', False):
+            warn("Command(s) not found:\n%s" \
                 % indent(unknown_commands))
+            show_commands(None, options.list_format, 1)
 
         # Generate remainder command and insert into commands, commands_to_run
         if remainder_command:
@@ -660,6 +724,11 @@ Remember that -f can be used to specify fabfile path, and use -h for help.""")
         if not commands_to_run and default:
             commands_to_run.append((default.name, [], {}, [], [], []))
 
+        # Initial password prompt, if requested
+        if options.initial_password_prompt:
+            prompt = "Initial value for env.password: "
+            state.env.password = getpass.getpass(prompt)
+
         if state.output.debug:
             names = ", ".join(x[0] for x in commands_to_run)
             print("Commands to run: %s" % names)
@@ -681,7 +750,7 @@ Remember that -f can be used to specify fabfile path, and use -h for help.""")
         raise
     except KeyboardInterrupt:
         if state.output.status:
-            print >> sys.stderr, "\nStopped."
+            sys.stderr.write("\nStopped.\n")
         sys.exit(1)
     except:
         sys.excepthook(*sys.exc_info())
